__doc__="""Cached SQL Methods
"""

import Shared.DC.ZRDB.DA
from Shared.DC.ZRDB import RDB
from Shared.DC.ZRDB.Results import Results
from Globals import DTMLFile
from Globals import InitializeClass
from AccessControl import ClassSecurityInfo
from AccessControl import getSecurityManager
from AccessControl.Permissions import change_database_methods
from App.Extensions import getBrain


from webdav.WriteLockInterface import WriteLockInterface
from OFS.Cache import Cacheable

from Products.ZSQLMethods.SQL import SQLConnectionIDs

import md5
import time

from logging import getLogger

LOG = getLogger('CachedSQL')

manage_addCZSQLMethodForm=DTMLFile('dtml/add',globals())

def manage_addCZSQLMethod(self,id,title,
                                connection_id,arguments,
                                template,
                                REQUEST=None):
    """ add a new Cached ZSQL Method """
    
    self._setObject(id,CachedSQL(id,title,connection_id,arguments,template))
    if REQUEST is not None:
        try: u=self.DestinationURL()
        except: u=REQUEST['URL1']
        if REQUEST.has_key('add_edit'):
            REQUEST.RESPONSE.redirect("%s/%s/manage_main" % (u,id))
        elif REQUEST.has_key('add_test'):
            REQUEST.RESPONSE.redirect("%s/%s/manage_testForm" % (u,id))
        else:
            REQUEST.RESPONSE.redirect("%s/manage_main" % u)
    return ''
        
_marker=[]

class CachedSQL(Shared.DC.ZRDB.DA.DA,Cacheable):
    meta_type='Z SQL Method Cached'
    security = ClassSecurityInfo()
    manage_options=Shared.DC.ZRDB.DA.DA.manage_options+Cacheable.manage_options 
    security.declareProtected(change_database_methods, 'manage')
    security.declareProtected(change_database_methods, 'manage_main')
    manage=manage_main=DTMLFile('dtml/edit', globals())
    manage_main._setName('manage_main')
    
    def __call__(self, REQUEST=None, __ick__=None, src__=0, test__=0, **kw):
        """Call the database method

        The arguments to the method should be passed via keyword
        arguments, or in a single mapping object. If no arguments are
        given, and if the method was invoked through the Web, then the
        method will try to acquire and use the Web REQUEST object as
        the argument mapping.

        The returned value is a sequence of record objects.
        
        too bad we have to copy here but we need special handling
        of instance values for caching
        """

        __traceback_supplement__ = (Shared.DC.ZRDB.DA.SQLMethodTracebackSupplement, self)

        if REQUEST is None:
            if kw: REQUEST=kw
            else:
                if hasattr(self, 'REQUEST'): REQUEST=self.REQUEST
                else: REQUEST={}


        if hasattr(self, 'aq_parent'):
            p=self.aq_parent
            if self._isBeingAccessedAsZClassDefinedInstanceMethod():
                p=p.aq_parent
        else: p=None

        argdata=self._argdata(REQUEST)
        
        query_key= {'connection':self.connection_hook,
                    'query':md5.new(self.template.read()).digest(),
                    'maxrows':str(self.max_rows_),
                    'argdata':dict(argdata)}        
        
        result=_marker

        if not src__:
            if self.ZCacheable_isCachingEnabled():
                result=self.ZCacheable_get(keywords=query_key,default=_marker)
                #if result is not _marker and not test:
                #    result
            
        if result is _marker or test__ or src__:
            starttime=time.time()
            # connection hook
            c = self.connection_id
            # for backwards compatability
            hk = self.connection_hook
            # go get the connection hook and call it
            if hk: c = getattr(self, hk)()
            try: dbc=getattr(self, c)
            except AttributeError:
                raise AttributeError, (
                  "The database connection cannot be found: %s" % (
                c))
            try: DB__=dbc()
            except: raise DatabaseError, (
                    '%s is not connected to a database' % self.id)
        
            argdata['sql_delimiter']='\0'
            argdata['sql_quote__']=dbc.sql_quote__
            security=getSecurityManager()
            security.addContext(self)
            try:
                try:     query=apply(self.template, (p,), argdata)
                except TypeError, msg:
                    msg = str(msg)
                    if find(msg,'client') >= 0:
                        raise NameError("'client' may not be used as an " +
                                    "argument name in this context")
                    else: raise
            finally: security.removeContext(self)
        
            if src__: return query
        
        if result is _marker:
            result=DB__.query(query,self.max_rows_)
            runtime=int((time.time()-starttime)*1000)
        
            if self.ZCacheable_isCachingEnabled() and runtime>self.cache_time_:
                LOG.info("Caching %r because runtime: %d ms" % (query_key,runtime))
                self.ZCacheable_set(result, keywords=query_key)
            elif self.ZCacheable_isCachingEnabled():
                LOG.info("Not caching %r because runtime: %d ms" % (query_key,runtime))
                    
        
        if hasattr(self, '_v_brain'): brain=self._v_brain
        else:
            brain=self._v_brain=getBrain(self.class_file_, self.class_name_)

        zc=self._zclass
        if zc is not None: zc=zc._zclass_

        if type(result) is type(''):
            f=StringIO()
            f.write(result)
            f.seek(0)
            result=RDB.File(f,brain,p, zc)
        else:
            result=Results(result, brain, p, zc)
        columns=result._searchable_result_columns()
        
        if test__ and columns != self._col: self._col=columns


        if test__: return query, result

        return result

        result=self._cached_result(query)
        
        if hasattr(self, '_v_brain'): brain=self._v_brain
        else:
            brain=self._v_brain=getBrain(self.class_file_, self.class_name_)

        zc=self._zclass
        if zc is not None: zc=zc._zclass_

        if type(result) is type(''):
            f=StringIO()
            f.write(result)
            f.seek(0)
            result=RDB.File(f,brain,p, zc)
        else:
            result=Results(result, brain, p, zc)
        columns=result._searchable_result_columns()
        if test__ and columns != self._col: self._col=columns

        # If run in test mode, return both the query and results so
        # that the template doesn't have to be rendered twice!
        if test__: return query, result

        return result
    
    def _cached_result(self,query):
        # we need to munge the incoming query key in the cache
        # so that the same request to a different db is returned
        query_key= {'connection':self.connection_hook,'query':md5.new(query).digest(),
                    'maxrows':str(self.max_rows_)}
        
        if self.ZCacheable_isCachingEnabled():
            result=self.ZCacheable_get(keywords=query_key,default=_marker)
            if result is _marker:
                starttime=time.time()
                # connection hook
                c = self.connection_id
                # for backwards compatability
                hk = self.connection_hook
                # go get the connection hook and call it
                if hk: c = getattr(self, hk)()
           
                try: dbc=getattr(self, c)
                except AttributeError:
                    raise AttributeError, (
                    "The database connection cannot be found: %s" % (
                    c))

                try: DB__=dbc()
                except: raise DatabaseError, (
                        '%s is not connected to a database' % self.id)

                result=DB__.query(query,self.max_rows_)
                runtime=int((time.time()-starttime)*1000)
                if runtime>self.cache_time_:
                    LOG.info("Caching %r because runtime: %d ms" % (query_key,runtime))
                    self.ZCacheable_set(result, keywords=query_key)
                else:
                    LOG.info("Not caching %r because runtime: %d ms" % (query_key,runtime))
                return result
        
        # connection hook
        c = self.connection_id
        # for backwards compatability
        hk = self.connection_hook
        # go get the connection hook and call it
        if hk: c = getattr(self, hk)()
        try: dbc=getattr(self, c)
        except AttributeError:
            raise AttributeError, (
            "The database connection cannot be found: %s" % (
            c))

        try: DB__=dbc()
        except: raise DatabaseError, (
            '%s is not connected to a database' % self.id)
        return DB__.query(query,self.max_rows_)